/*
 *	Expat hash table - based on hashtable.c
 *
 *	Copyright (c) 1998,2003 Adobe Systems Inc.
 *	All Rights Reserved
 *
 *	The Initial Developer of the Original Code is James Clark.
 *	Portions created by James Clark are Copyright (C) 1998
 *	James Clark. All Rights Reserved.
 *
 *	XPHashtable.cpp
 *
 *		Implementation of C++ wrapper to Expat hash table.
 */

#include <string.h>
#include "XPString.h"
#include "XPHashtable.h"
#include "acroassert.h"

#include "ASSafeAPIs.h" // srdas: Unsafe to Safe replacements.

#ifdef ACROASSERT
#define ACROASSERTRETURN(x)         if (!(x)) return (void)ACROASSERT(x)
#define ACROASSERTRETURNVALUE(x, y) if (!(x)) return       ACROASSERT(x), y
#endif

#ifndef NULL
#define NULL    0
#endif

#ifdef WIN_ENV
/* mradwin: debug heap allocations with line number, etc. */
#include <crtdbg.h>
#endif

#ifdef _DEBUG
	#ifdef WIN_ENV
		#define DEBUG_NEW (_NORMAL_BLOCK, __FILE__, __LINE__)
	#else
		#define DEBUG_NEW
	#endif // WIN_ENV
#else
	#define DEBUG_NEW
#endif

const int INIT_SIZE = 64;

XPHashtable::XPHashtable(unsigned long initSize, bool copyStrings)
{
	#define IsAPowerOfTwo(x) ( !( (x) & ((x)-1) ) )
	ACROASSERT(IsAPowerOfTwo(initSize));

	m_initSize = (initSize > 0) ? initSize : INIT_SIZE;
	m_size = 0;
	m_usedLim = 0;
	m_used = 0;
	m_v = 0;
	m_usesCopiedStrings = copyStrings;
}

// Strings were duplicated, values not, so don't delete everything.
XPHashtable::~XPHashtable()
{
	if (m_usesCopiedStrings)
	{
		for (unsigned int i = 0; i < m_size; i++)
		{
			char *cmc = reinterpret_cast<char *>( m_v[2*i] );
			delete [] cmc;
		}
	}
	delete [] m_v;
}

unsigned long XPHashtable::hash(const char *s, unsigned long len)
{
	unsigned long h = 0;
	ACROASSERTRETURNVALUE(s, h);
	while (*s && len--)
		h = (h << 5) + h + (unsigned long)*s++;
	return h;
}

// Returned index is either index of found key or where key should go.
bool XPHashtable::lookup(const char* key, unsigned long *index, unsigned long len) const
{
	ACROASSERTRETURNVALUE(key, false);
	bool foundKey = false;
	unsigned long h = hash(key, len);
	if (m_size == 0)
	{
		*index = h & (m_initSize - 1);
		return false;
	}
	unsigned long i = h & (m_size - 1);
	while (m_v[2*i] != NULL)
	{
		// strcmp
		const char *c1 = (const char *) m_v[2*i];
		const char *c2 = key;
		unsigned long tmp_len = len;
		if (tmp_len != (unsigned long)(-1))
			tmp_len++;
		for(; tmp_len; tmp_len--)
		{
			if (*c1 != *c2)
			{
				if (*c1 == 0 && tmp_len == 1)
					tmp_len = 0;
				break;
			}
			if (*c1 == 0)
			{
				tmp_len = 0;
				break;
			}
			c1++; 
			c2++;
		}

		if (tmp_len == 0) // string match?
		{
			foundKey = true;
			break;
		}
		if (i == 0)
			i = m_size - 1;
		else
			--i;
	}
	*index = i;
	return foundKey;
}

void *XPHashtable::get(const char* key, unsigned long len) const
{
	ACROASSERTRETURNVALUE(key, (void *)0);

	unsigned long i;
	return (lookup(key, &i, len)) ? m_v[2*i + 1] : NULL;
}

void *XPHashtable::get(const XPString& key) const
{ 
	return get(key.c_str()); 
}

void *XPHashtable::put(const char* key, void *value)
{
	ACROASSERTRETURNVALUE(key, (void *)0);

	bool found = false;
	unsigned long i;
	if (m_size == 0)
	{
		m_v = new DEBUG_NEW void *[2*m_initSize];
		if (!m_v)
			return NULL;
		memset(m_v, '\0', (2 * m_initSize) * sizeof(m_v[0]));
		for (i = 0; i < 2 * m_initSize; i++)
			m_v[i] = NULL;
		m_size = m_initSize;
		m_usedLim = m_initSize / 2;
		i = hash(key) & (m_size - 1);
	}
	else
	{
		found = lookup(key, &i);
		if (!found && m_used == m_usedLim)
		{
			/* check for overflow */
			unsigned long newSize = m_size * 2;
			void **newV = new DEBUG_NEW void *[2*newSize];
			if (!newV)
				return NULL;
			memset(newV, '\0', (2 * newSize) * sizeof(newV[0]));
			for (i = 0; i < m_size; i++)
			{
				if (m_v[2*i] != NULL)
				{
					unsigned long j;
					j = hash((const char *) m_v[2*i]) & (newSize - 1);
					while(newV[2*j] != NULL)
					{
						if (j == 0)
							j = newSize - 1;
						else
							--j;
					}
					newV[2*j] = m_v[2*i];
					newV[2*j + 1] = m_v[2*i + 1];
				}
			}
			delete [] m_v;
			m_v = newV;
			m_size = newSize;
			m_usedLim = newSize / 2;
			lookup(key, &i);
		}
	}
	if (!found)
	{
		long keylengthplus1 = ASstrnlen_unsafe(key) + 1;
		char *p;
		if (m_usesCopiedStrings == true)
		{
			p = new DEBUG_NEW char[keylengthplus1];
			memcpy(p, key, keylengthplus1);
		}
		else
		{
			// use the pointer to the key passed in
			// IT MUST BE PERMANENTLY MAINTAINED (at least as long as the hashtable is)
			p = const_cast<char *>(key);
		}
		m_v[2*i] = (void *) p;
		m_used++;
	}
	void *oldValue = m_v[2*i+1];
	m_v[2*i+1] = value;
	return oldValue;
}

void *XPHashtable::put(const XPString& key, void *value)  // return old value
{ 
	ACROASSERTRETURNVALUE(m_usesCopiedStrings == true, (void *)0);
	return put(key.c_str(), value);
}

